
/************************************************************/
/* exact determination of mq and Zq once the Vijk are fixed */
/************************************************************/

void set_mZ(char q)
{
    std::string t[3]={"11","1r","rr"};
    // u: lower bound on angles 1q1, 1qr and rqr in any FM-triangle
    std::array<I,3> u;
    for (int i=0;i<3;i++)
        u[i]=lower(asin(min(val(t[i][0])/(val(q)+val(t[i][0])+I(2)*r),val(t[i][1])/(val(q)+val(t[i][1])+I(2)*r))));
    // a: angles 1q1, 1qr and rqr in a tight triangle
    std::array<I,3> a;
    for (int i=0;i<3;i++)
        a[i]=angle(val(t[i][0])+val(t[i][1]),val(q)+val(t[i][0]),val(q)+val(t[i][1]));
    // ranges through coronas k such that sum of angle may be <=2*pi
    int k[3]={0,0,0};
    I z=0; // sum of angles in corona k
    int i=0; // index of the "digit" of k to increment
    double mq=0;
    while (i<3)
    {
        // increment corona k
        i=0;
        while (i<3 && lower(z+u[i])>upper(pi_twice<I>()))
        {
            z-=I(k[i])*u[i];
            k[i]=0;
            i+=1;
        }
        if (i==3) break; // all the coronas have been considered -> finished
        k[i]+=1;
        z+=u[i];
        // skip q-config qqqqqq (positivity ensured by Vqqq>=0)
        if (q=='1' && k[0]==6 && k[1]==0 && k[2]==0) continue;
        if (q=='r' && k[0]==0 && k[1]==0 && k[2]==6) continue;
        // skip non-coronable k
        if (k[1]%2!=0 || (k[0]*k[2]!=0 && k[1]==0) || (k[0]+k[1]+k[2]<3)) continue;
        // compute the coefficient of mq
        I diff_angle=abs(pi_twice<I>()-I(k[0])*a[0]-I(k[1])*a[1]-I(k[2])*a[2]);
        // compute the sum of vertex base potentials
        I base_pot;
        if (q=='1') base_pot=I(k[0])*V111+I(k[1])*V11r+I(k[2])*Vr1r;
        if (q=='r') base_pot=I(k[0])*V1r1+I(k[1])*V1rr+I(k[2])*Vrrr;
        // WARNING: a q-corona may be not detected if zero_in(diff_angle) but upper(diff_angle) too large.        
        if (upper(diff_angle)<min_diff_angle) // special case: corona
        {
            // vertex positivity should have been ensured by the choice of the Vijk's (function set_xy_corona)
            if (verbose) printf("%c-corona %d,%d,%d\n",q,k[0],k[1],k[2]);
            if (upper(base_pot)<0) throw("error");
        }
        else // normal case (no corona)
        {        
            // updates the lower bound on mq
            double mk=upper(-base_pot/diff_angle);
            if (mk>mq)
            {
                mq=mk;
                if (verbose) printf("%c-corona [%d,%d,%d] increases m%c to %f\n",q,k[0],k[1],k[2],q,mq);
            }
        }
    }
    q=='1' ? m1=mq : mr=mq;
    // ceiling Zq (per radian)
    I Zq=-(q=='1' ? boost::numeric::min(boost::numeric::min(V111/a[0],V11r/a[1]),Vr1r/a[2]) : boost::numeric::min(boost::numeric::min(V1r1/a[0],V1rr/a[1]),Vrrr/a[2]));
    mq=std::max(mq,upper(Zq));
    // update m1/Z1 or mr/Zr
    if (q=='1') {m1=mq; Z1=upper(Zq);}
    else  {mr=mq; Zr=upper(Zq);}
}

/***************************************************/
/* exact determination of eps once m1 and mr fixed */
/***************************************************/

bool aux_test_eps(I ra,I rb,I rc, I eps)
{
    I a=rb+rc+I(0,upper(eps));
    I b=rc+ra+I(0,upper(eps));
    I c=ra+rb+I(0,upper(eps));
    I de=((square(a)+square(b)-square(c))*square(rb)+(square(a)+square(c)-square(b))*square(rc))/(I(2)*a)+a*((square(b)+square(c)-square(a))*delta-I(2)*square(ra))/I(2);
    I du=I(2)*a*(ra==I(1)? m1 : mr)+abs(square(a)+square(b)-square(c))/a*(rb==I(1) ? m1 : mr)+abs(square(a)+square(c)-square(b))/a*(rc==I(1) ? m1 : mr);
    return (lower(de)>=upper(du));
}

bool test_eps(I eps)
{
    I u=I(1);
    return (
       aux_test_eps(u,u,u,eps)
    && aux_test_eps(u,u,r,eps)
    && aux_test_eps(u,r,r,eps)
    && aux_test_eps(r,u,u,eps)
    && aux_test_eps(r,u,r,eps)
    && aux_test_eps(r,r,r,eps));
}

void set_eps()
{
    I eps_sup=I(0,1); // will be such that ok for lower(eps_sup)-tight triangles
    std::pair<I,I> half;
    while (width(eps_sup)>1e-10)
    {
        half=bisect(eps_sup);
        if (test_eps(half.first)) eps_sup=half.second;
        else eps_sup=half.first;
    }
    eps=lower(eps_sup);
}

/**************************************************/
/* choose parameters for the potential of edge ij */
/**************************************************/

// find by dichotomy the length of edge ij such that the support disc of the triangle with Tij1 with contacts i1 and j1 is centered on the edge ij
double aux_L(char i, char j, char k)
{
    I ri=val(i);
    I rj=val(j);
    I rk=val(k);
    I x1=ri+rj; // edge ij in the tight triangle
    I x2=sqrt(ri*(ri+I(2)*rk))+sqrt(rj*(rj+I(2)*rk)); // edge ij in the stretched triangle
    I x;
    I rs;
    while (lower(x2)-upper(x1)>1e-5)
    {
        x=(x1+x2)/I(2);
        rs=radius(rj+rk,ri+rk,x,ri,rj,rk);
        if (asap) //find x such that the support disc is centered on the edge ij
        {
            if (median(dist(x,rj+rk,ri+rk,rk,ri,rj,rs))>0) x1=x; else x2=x;
        }
        else // set x so that the edge pot. acts slightly before Uv reaches E
        {
            if (median(E(rj+rk,ri+rk,x,ri,rj,rk)-Uv(rj+rk,ri+rk,x,i,j,k))>1e-2) x1=x; else x2=x;
        }
    }
    return median(x);
}

double aux_Q(char i, char j, char k)
{
    I ri=val(i);
    I rj=val(j);
    I rk=val(k);
    I x=sqrt(ri*(ri+I(2)*rk))+sqrt(rj*(rj+I(2)*rk)); // length of stretched edge ij
    // set q such that U is slightly less than E for stretched triangle
    // in the stretched triangle, the support disc has radius rk and is at distance -rk from the stretched edge
    I q=(E(rj+rk,ri+rk,x,ri,rj,rk)-Uv(rj+rk,ri+rk,x,i,j,k)-0.01)/(-rk);
    return std::max(0.0,median(q));
}

void set_LQ(char i, char j)
{
    // the edge potential will act beyond l
    double l=std::min(aux_L(i,j,'1'),aux_L(i,j,'r'));
    // with intensity q
    double q=std::max(aux_Q(i,j,'1'),aux_Q(i,j,'r'));
    if (i=='1' && j=='1') {l11=l; q11=q;}
    else if (i!=j) {l1r=l; q1r=q;}
    else {lrr=l; qrr=q;}
}

