
// naive conversion of an interval x to a gmp integer
mpz_class convert(I x)
{
    // upper/ceil -> any exactly valid parameter is in the polytope (extensive)
    // lower/floor -> any point in the polytope is an exactly valid parameter (restrictive)
    mpf_class y=lower(x);
    y=floor(y*pow(2,20));
    return mpz_class(y);
}

// compute the barycenter of the vertices of a polytope P
PPL::Generator center(PPL::Polyhedron* P)
{
    PPL::Generator_System gs = (*P).generators();
    mpz_class Xc=0,Yc=0,M1c=0,Mrc=0,dc=1;
    int k=0;
    for(PPL::Generator_System::const_iterator it = gs.begin(); it != gs.end(); it++)
    {
        const PPL::Generator g = *it;
        if (!g.is_point()) continue;
        Xc=Xc*g.divisor()+g.coefficient(X)*dc;
        Yc=Yc*g.divisor()+g.coefficient(Y)*dc;
        M1c=M1c*g.divisor()+g.coefficient(M1)*dc;
        Mrc=Mrc*g.divisor()+g.coefficient(Mr)*dc;
        dc=dc*g.divisor();
        k+=1;
    }
    dc=dc*k;
    PPL::Generator c=PPL::point(Xc*X+Yc*Y+M1c*M1+Mrc*Mr,dc);    
    return c;
}

/******************************************************/
/* constraints which should ensure eps>=0 (non exact) */
/******************************************************/

// auxiliary function
void check(I ra, I rb, I rc)
{
    I a=rb+rc+I(0,min_eps);
    I b=rc+ra+I(0,min_eps);
    I c=ra+rb+I(0,min_eps);
    I dE=((square(a)+square(b)-square(c))*square(rb)+(square(a)+square(c)-square(b))*square(rc))/(I(2)*a)+a*((square(b)+square(c)-square(a))*delta-I(2)*square(ra))/I(2);
    I dU[2]={I(0),I(0)}; // coeffs de m1 et mr dans dU
    dU[lower(ra)==1?0:1]+=I(2)*a;
    dU[lower(rb)==1?0:1]+=abs(square(a)+square(b)-square(c))/a;
    dU[lower(rc)==1?0:1]+=abs(square(a)+square(c)-square(b))/a;
    // dE-dU>=0 ?

//    printf("%f,%f,%f\n",lower(dE),lower(-dU[0]),lower(-dU[1])); //OK

    // approximation de la direction réelle dE-dU par une direction entière
    std::array<mpz_class,3> ieq;
    ieq[0]=convert(dE); // dE
    ieq[1]=convert(-dU[0]); // coeff m1 dans -dU
    ieq[2]=convert(-dU[1]); // coeff mr dans -dU
//    printf("1, 0, 0, %f, %f\n",(double)ieq2[1]/(double)ieq2[0],(double)ieq2[2]/(double)ieq2[0]); //OK
//    printf("%f=%f=%f\n\n",lower(-dU[1])/lower(dE),ieq[2]/ieq[0],ieq2[2]/ieq2[0]); //ok

    ieqs.insert(ieq[0]+ieq[1]*M1+ieq[2]*Mr>=0);
}

// ineqs on (x, y, m1, mr) to ensure eps>=0
void add_epsilon_constraints()
{
    I u=1;
    check(u,u,u);
    check(u,u,r);
    check(u,r,u);
    check(r,u,u);
    check(u,r,r);
    check(r,u,r);
    check(r,r,u);
    check(r,r,r);
}

/*****************************************************/
/* constraints which should ensure vertex-inequality */
/*****************************************************/

void add_vertex_positivity_constraints(char q)
{
    int nieqs=0;
    std::string t[3]={"11","1r","rr"};
    // u: lower bound on angles 1q1, 1qr and rqr in any FM-triangle
    I* u=(I*)calloc(3,sizeof(I));
    for (int i=0;i<3;i++)
        u[i]=lower(asin(min(val(t[i][0])/(val(q)+val(t[i][0])+I(2)*r),val(t[i][1])/(val(q)+val(t[i][1])+I(2)*r))));
    // a: angles 1q1, 1qr and rqr in a tight triangle
    I* a=(I*)calloc(3,sizeof(I));
    for (int i=0;i<3;i++)
        a[i]=angle(val(t[i][0])+val(t[i][1]),val(q)+val(t[i][0]),val(q)+val(t[i][1]));
    // ranges through coronas k such that sum of angle may be less or equal 2*pi
    int k[3]={0,0,0};
    I z=0; // sum of angles in corona k
    int i=0; // index of the "digit" of k to increment
    I mq=I(0);
    while (i<3)
    {
        // increment corona k
        i=0;
        while (i<3 && lower(z+u[i])>upper(pi_twice<I>()))
        {
            z-=I(k[i])*u[i];
            k[i]=0;
            i+=1;
        }
        if (i==3) break; // all the coronas have been considered -> finished
        k[i]+=1;
        z+=u[i];
        // skip q-config qqqqqq (positivity ensured by Vqqq>=0)
        if (q=='1' && k[0]==6 && k[1]==0 && k[2]==0) continue;
        if (q=='r' && k[0]==0 && k[1]==0 && k[2]==6) continue;
        // skip non-coronable k
        if (k[1]%2!=0 || (k[0]*k[2]!=0 && k[1]==0) || (k[0]+k[1]+k[2]<3)) continue;
        // inequality that will be added
        std::array<mpz_class,5> ieq;
        // angle difference
        I diff_angle=abs(pi_twice<I>()-I(k[0])*a[0]-I(k[1])*a[1]-I(k[2])*a[2]);
        // q-corona detected ?
        if (upper(diff_angle)<min_diff_angle)
        {
            if (q=='1')
            {
                exists_large_corona=true;
                large_corona[0]=k[0];
                large_corona[1]=k[1];
                large_corona[2]=k[2];
            }
            else
            {
                exists_small_corona=true;
                small_corona[0]=k[0];
                small_corona[1]=k[1];
                small_corona[2]=k[2];
            }
            //if (verbose) 
            printf("%c-corona [%d,%d,%d]\n",q,k[0],k[1],k[2]);
            continue; // vertex inequality will be ensured by the choices of x=V11r and y=V1rr -> does not refine the polytope
        }
        // build diophantine inequality ieq (approximation since exact coefficients are real numbers)
        if (q=='1')
        {
            // coeff constant : k0*E111/3+k2*E1rr
            ieq[0]=convert(I(k[0])*E111/I(3)+I(k[2])*E1rr);
            // coeff x
            ieq[1]=convert(-I(2*k[2]));
            // coeff y
            ieq[2]=convert(I(k[1]));
            // coeff m1
            ieq[3]=convert(diff_angle);
            // coeff mr
            ieq[4]=0;
        }
        else
        {
            // coeff constant : k0*E11r+k2*Errr/3
            ieq[0]=convert(I(k[0])*E11r+I(k[2])*Errr/I(3));
            // coeff x
            ieq[1]=convert(I(k[1]));
            // coeff y
            ieq[2]=convert(-I(2*k[0]));
            // coeff m1
            ieq[3]=0;
            // coeff mr
            ieq[4]=convert(diff_angle);
        }
        // add inegality to constrain system
        ieqs.insert(ieq[0]+ieq[1]*X+ieq[2]*Y+ieq[3]*M1+ieq[4]*Mr>=0);
        nieqs+=1;        
    }
    if (verbose) printf("%d inequalities for q=%c\n",nieqs,q);
    free(u);
    free(a);
}

/***********************************************************/
/* the art of choosing "good" parameters x=V1rr and y=V11r */
/***********************************************************/

void set_xy_generic(int strategy=0)
{
    // 4D polytope defined by the constraints
    PPL::C_Polyhedron P(ieqs);

    // quite arbitrary selection of a point in the polytope

    bool b;
    // first forces Mr/M1 to be equal to (max value of Mr)/(max value of M1)
    PPL::Coefficient n1,d1,nr,dr;
    P.maximize(Mr,nr,dr,b);
    P.maximize(M1,n1,d1,b);
    P.refine_with_constraint(n1*dr*Mr==nr*d1*M1);

    // then choose a point g in the obtained polytope
    PPL::Generator g;
    if (upper(r)<0.55) P.minimize(M1,n1,d1,b,g); //minimizes M1 (or, equivalently, Mr since the ratio Mr/M1 is fixed)
    else g=center(&P); //polytope center

    // it defines x and y
    mpq_class x=g.coefficient(X);
    mpq_class y=g.coefficient(Y);
    x=x/g.divisor();
    y=y/g.divisor();

    // set the two parameters V1rr=x and V11r=y
    V1rr=I(x.get_d());
    V11r=I(y.get_d());

    // deduce the other Vijk's
    V111=E111/I(3);
    Vrrr=Errr/I(3);
    V1r1=E11r-I(2)*V11r;
    Vr1r=E1rr-I(2)*V1rr;
}

// if there are (exact) coronas, x and y must be chosen so that the vertex inequality is theoretically ensured
// it means: we choose x and deduce y (or conversely)
// two coronas never yield independent equations, so it does not matter which corona we consider
void set_xy_corona()
{
    // theoretically ensure vertex positivity for (eventual) exact coronas
    if (exists_large_corona && exists_small_corona) // 2 coronas (triangulated packing) -> solve linear system of 2 eqs on x and y
    {
        int a=2*large_corona[2];
        int b=-large_corona[1];
        int c=-small_corona[1];
        int d=2*small_corona[0];
        int det=a*d-b*c;
        I u=I(large_corona[0])*E111/I(3)+I(large_corona[2])*E1rr;
        I v=I(small_corona[0])*E11r+I(small_corona[2])*Errr/I(3);
        V1rr=(I(d)*u-I(b)*v)/I(det);
        V11r=(-I(c)*u+I(a)*v)/I(det);
    }
    else if (exists_large_corona)
    {
        // computes y=V11r as a function of x=V1rr if possible, conversely otherwise
        if (large_corona[1]!=0) // y as a function of x, updates V1r1 accordingly
            V11r=(I(large_corona[0])*E111/I(3)+I(large_corona[2])*E1rr-I(2*large_corona[2])*V1rr)/I(-large_corona[1]);
        else // x as a function of y, updates Vr1r accordingly
            V1rr=(I(large_corona[0])*E111/I(3)+I(large_corona[2])*E1rr)/I(2*large_corona[2]);
    }
    else // r-corona toute seule : n'arrive que pour rrrrr i.e. r=0.70...
    {
        if (small_corona[1]!=0)
            V1rr=(I(small_corona[0])*E11r+I(small_corona[2])*Errr/I(3)-I(2*small_corona[0])*V11r)/I(-small_corona[1]);
        else
            V11r=(I(small_corona[0])*E11r+I(small_corona[2])*Errr/I(3)+I(small_corona[1])*V1rr)/I(2*small_corona[0]);
    }
    // updates V1r1 and Vr1r accordingly
    V1r1=E11r-I(2)*V11r;
    Vr1r=E1rr-I(2)*V1rr;

}

